from __future__ import annotations

from contextlib import asynccontextmanager
from contextvars import ContextVar
from datetime import datetime, timedelta
import logging
from pathlib import Path, PurePosixPath
from typing import AsyncIterable, AsyncIterator, Optional, Union

from httpx import AsyncClient, HTTPStatusError, Request
from tenacity import retry, stop_after_attempt, wait_random
from tqdm.asyncio import tqdm

from nasa_csda.config import Settings
from nasa_csda.models.link import DownloadLink
from nasa_csda.models.item_collection import CSDAItemCollection
from nasa_csda.models.search import CSDASearch
from nasa_csda.transport import RetryableTransport

logger = logging.getLogger(__name__)
_session: ContextVar[AsyncClient] = ContextVar("session")
_client: ContextVar[Client] = ContextVar("client")


class CSDAClientError(Exception):
    pass


class Client(object):
    def __init__(self, config: Settings) -> None:
        self.config = config
        self._access_token: Optional[str] = None
        self._refresh_token: Optional[str] = None
        self._expiration: Optional[datetime] = None

    @asynccontextmanager
    async def stream_context(self) -> AsyncIterator[Client]:
        token = _client.set(self)
        async with self.session():
            yield self
        _client.reset(token)

    @property
    def current_session(self) -> AsyncClient:
        try:
            return _session.get()
        except LookupError:
            raise CSDAClientError("Client method must be used inside a session context.")

    @classmethod
    def current_client(self) -> Client:
        try:
            return _client.get()
        except LookupError:
            raise CSDAClientError("Client method must be used inside a streaming context.")

    @asynccontextmanager
    async def session(self) -> AsyncIterator[AsyncClient]:
        config = self.config
        transport = RetryableTransport(config)

        async def set_auth(request: Request):
            if str(self.config.api) in str(request.url):
                request.headers.setdefault("Authorization", f"Bearer {await self.get_token()}")

        async with AsyncClient(
            transport=transport,
            base_url=str(self.config.api),
            timeout=None,
            event_hooks={
                "request": [set_auth],
            },
        ) as session:
            if self._access_token is None:
                await self._login(session)
            token = _session.set(session)
            yield session
            _session.reset(token)

    async def get_token(self) -> str:
        session = self.current_session
        assert self._access_token, "Login failed"
        if not self._expiration or self._expiration < datetime.now():
            await self._refresh(session)
        return self._access_token

    async def search(self, *queries: CSDASearch) -> AsyncIterator[CSDAItemCollection]:
        page_size = self.config.search_page_size
        session = self.current_session
        token: Optional[str] = None
        item_count = 0

        # https://github.com/encode/httpx/discussions/2056
        @retry(stop=stop_after_attempt(10), wait=wait_random(1, 5))
        async def execute_query(query):
            resp = await session.post("stac/search", json=query)
            if resp.status_code == 401:
                # The token is invalid so log in again
                await self._login(session)
            if resp.status_code != 200:
                raise ValueError(resp.content)
            return resp

        for query in queries:
            while True:
                query_json = query.model_copy(update={"token": token, "limit": page_size}).model_dump(
                    mode="json",
                    by_alias=True,
                    exclude_none=True,
                )
                resp = await execute_query(query_json)
                items = CSDAItemCollection.model_validate_json(resp.content)
                yield items
                token = items.next_token
                item_count += len(items.features)
                if token is None:
                    break

    async def download_links(
        self,
        iterator: AsyncIterable[CSDAItemCollection],
        /,
    ) -> AsyncIterator[DownloadLink]:
        async for item_collection in iterator:
            for item in item_collection.features:
                for asset in item.assets.values():
                    try:
                        url = DownloadLink.parse_url(f"{self.config.api}{asset.href.lstrip('/')}")
                    except Exception:
                        logger.exception(f"Could not parse {asset.href}")
                    yield url

    async def download(
        self,
        links: AsyncIterable[Union[str, DownloadLink]],
        prefix: str,
        overwrite: bool = False,
    ) -> AsyncIterator[Optional[str]]:

        async for link in links:
            yield await self.download_file(link, prefix=prefix, overwrite=overwrite, progress=self.config.download_progress)

    async def download_file(
        self,
        url: Union[str, DownloadLink],
        prefix: str,
        *,
        overwrite: bool = False,
        progress: bool = False,
    ) -> Optional[str]:
        session = self.current_session
        if not isinstance(url, DownloadLink):
            url = DownloadLink.parse_url(url)
        prefix_path = Path(PurePosixPath(prefix.format_map(url.model_dump())))
        destination = prefix_path / str(url).split("/")[-1]
        if destination.exists() and not overwrite:
            return str(destination)
        prefix_path.mkdir(parents=True, exist_ok=True)
        try:
            with destination.open(mode="wb") as f:
                async with session.stream("GET", str(url), follow_redirects=True) as resp:
                    if resp.status_code == 404:
                        logger.warning(f"Missing file at {url}")
                        return None
                    try:
                        resp.raise_for_status()
                    except HTTPStatusError:
                        logger.error((await resp.aread()).decode())
                    total = None
                    if "content-length" in resp.headers:
                        total = int(resp.headers["content-length"])
                    prog = tqdm(
                        unit="b",
                        unit_scale=True,
                        desc=f"{destination.name:30}",
                        leave=False,
                        total=total,
                        delay=0.5,
                        disable=not progress,
                    )
                    with prog as bar:
                        async for chunk in resp.aiter_bytes():
                            bar.update(len(chunk))
                            f.write(chunk)
        except BaseException:
            # delete partial file on any exception
            destination.unlink(missing_ok=True)
            raise
        return str(destination)

    async def _login(self, session: AsyncClient) -> None:
        config = self.config
        data = {
            "AuthFlow": "USER_PASSWORD_AUTH",
            "ClientId": config.cognito_client_id,
            "AuthParameters": {
                "PASSWORD": self.config.password.get_secret_value(),
                "USERNAME": self.config.username,
            },
        }
        headers = {
            "Content-Type": "application/x-amz-json-1.1",
            "X-Amz-Target": "AWSCognitoIdentityProviderService.InitiateAuth",
        }
        resp = await session.post(self.config.cognito_endpoint, json=data, headers=headers)
        if resp.status_code != 200:
            raise CSDAClientError("Authentication failure")
        auth = resp.json()["AuthenticationResult"]
        self._expiration = datetime.now() + timedelta(seconds=auth["ExpiresIn"]) - timedelta(minutes=5)
        self._access_token = auth["AccessToken"]
        self._refresh_token = auth["RefreshToken"]

    async def _refresh(self, session: AsyncClient) -> None:
        config = self.config

        # If the refresh token has expired, we try logging in again.
        data = {
            "AuthFlow": "REFRESH_TOKEN_AUTH",
            "ClientId": config.cognito_client_id,
            "AuthParameters": {
                "REFRESH_TOKEN": self._refresh_token,
            },
        }
        headers = {
            "Content-Type": "application/x-amz-json-1.1",
            "X-Amz-Target": "AWSCognitoIdentityProviderService.InitiateAuth",
        }
        resp = await session.post(
            "https://cognito-idp.us-east-1.amazonaws.com/",
            json=data,
            headers=headers,
        )
        if resp.status_code != 200:
            await self._login(session)

        assert self._access_token, "Login failed"
        auth = resp.json()["AuthenticationResult"]
        self._access_token = auth["AccessToken"]



from __future__ import annotations

from datetime import datetime as datetime_, timezone
from enum import Enum
from typing import Any, Iterator, Literal, Optional, Union

from dateutil.relativedelta import relativedelta
from pydantic import BaseModel, Field
from stac_pydantic.api.search import ExtendedSearch


class Operator(str, Enum):
    EQ = "="
    AND = "and"
    OR = "or"
    LIKE = "like"
    IN = "in"
    T_INTERSECTS = "t_intersects"  # temporal intersection
    S_INTERSECTS = "s_intersects"  # spatial intersection
    S_WITHIN = "s_within"  # geometry within
    S_CONTAINS = "s_contains"  # geometry contains
    A_CONTAINS = "a_contains"  # array operator contains
    A_OVERLAPS = "a_overlaps"


class GeometryType(str, Enum):
    POLYGON = "Polygon"
    POINT = "Point"


class PropertyQuery(BaseModel):
    property: str


class GeometryTypeQuery(BaseModel):
    type: GeometryType


class IntervalQuery(BaseModel):
    interval: list[str]


class DatetimeQuery(BaseModel):
    timestamp: datetime_


class CoordinatesQuery(BaseModel):
    coordinates: Any


class FilterExpression(BaseModel):
    """
    https://github.com/stac-api-extensions/filter
    https://docs.up42.com/developers/api-assets/stac-cql
    """

    op: Operator
    args: list[
        Union[
            PropertyQuery,
            IntervalQuery,
            FilterExpression,
            DatetimeQuery,
            CoordinatesQuery,
            GeometryTypeQuery,
            int,
            float,
            str,
            list,
            datetime_,
        ]
    ]


class CSDASearch(ExtendedSearch):
    token: Optional[str] = None
    filter: Optional[FilterExpression] = None
    filter_lang: Literal["cql2-json"] = Field("cql2-json", alias="filter-lang")

    @property
    def start_end(self) -> tuple[datetime_, datetime_]:
        start = datetime_(2019, 10, 1) if self.start_date is None else self.start_date
        end = datetime_.now() if self.end_date is None else self.end_date
        start = start.replace(tzinfo=timezone.utc) if start.tzinfo is None else start
        end = end.replace(tzinfo=timezone.utc) if end.tzinfo is None else end
        return start, end

    def split_by_datetime(self) -> Iterator[CSDASearch]:
        start, end = self.start_end
        month = start.replace(day=1, hour=0, minute=0, second=0, microsecond=0)
        while month <= end:
            month_start = max(start, month).isoformat()
            month_end = min(end, month + relativedelta(months=1)).isoformat()
            yield self.copy(update={"datetime": f"{month_start}/{month_end}"})
            month += relativedelta(months=1)

    def split_by_product(self) -> Iterator[CSDASearch]:
        if self.filter and self.filter.op == Operator.IN and self.filter.args[0] == PropertyQuery(property="item_type"):
            for product in self.filter.args[1]:  # type: ignore
                yield self.copy(update={"filter": FilterExpression(op=Operator.EQ, args=[self.filter.args[0], product])})
        else:
            yield self

    def split(self) -> Iterator[CSDASearch]:
        for q in self.split_by_datetime():
            for p in self.split_by_product():
                yield p

    @classmethod
    def build_query(
        cls,
        start_date: datetime_,
        end_date: datetime_,
        min_latitude: float,
        max_latitude: float,
        min_longitude: float,
        max_longitude: float,
        products: str,
    ) -> CSDASearch:
        if start_date.tzinfo is None:
            start_date = start_date.replace(tzinfo=timezone.utc)
        if end_date.tzinfo is None:
            end_date = end_date.replace(tzinfo=timezone.utc)

        field_filter = None
        if products:
            field_filter = {
                "op": "in",
                "args": [
                    {"property": "item_type"},
                    products.split(","),
                ],
            }

        datetime = f"{start_date.isoformat()}/{end_date.isoformat()}"
        return CSDASearch(
            bbox=[
                min_longitude,
                min_latitude,
                max_longitude,
                max_latitude,
            ],
            datetime=datetime,
            filter=field_filter,
        )



from datetime import datetime as datetime_, timezone
import re

from pydantic import BaseModel

_link_re = re.compile(
    r"^.*/download"
    r"/(?P<collection>[^/]+)"
    r"/(?P<datetime>\d{4}-\d{2}-\d{2}T\d{2}-\d{2}-\d{2})"
    r"_(?P<receiver>FM\d+)"
    r"[^/]*"
    r"_(?P<product>[^/_]+)"
    r".*$"
)


class DownloadLink(BaseModel):
    url: str
    collection: str
    datetime: datetime_
    receiver: str
    product: str

    @classmethod
    def parse_url(cls, url: str) -> "DownloadLink":
        m = _link_re.match(url)
        if m is None:
            raise ValueError(f"Could not parse {url}")
        data = m.groupdict()
        data["datetime"] = datetime_.strptime(data["datetime"], "%Y-%m-%dT%H-%M-%S").replace(tzinfo=timezone.utc)
        return cls(url=url, **data)

    def __str__(self) -> str:
        return self.url

    @property
    def file(self) -> str:
        return self.url.split("/")[-1]



from pydantic import BaseModel, ConfigDict, Field

from nasa_csda.models.search import CSDASearch


class BulkDownload(BaseModel):
    model_config = ConfigDict(extra="ignore")

    queries: list[CSDASearch] = Field(..., min_length=1)



from collections.abc import MutableMapping
import logging
from typing import AsyncIterable, AsyncIterator, Optional

from aiostream.core import pipable_operator, streamcontext
from aiostream.stream import advanced
from cachetools import LRUCache

from nasa_csda.client import Client
from nasa_csda.models.link import DownloadLink
from nasa_csda.models.item_collection import CSDAItemCollection
from nasa_csda.models.search import CSDASearch

logger = logging.getLogger(__name__)


@pipable_operator
async def search(
    query: AsyncIterable[CSDASearch],
    client: Client,
    task_limit: Optional[int] = None,
) -> AsyncIterator[CSDAItemCollection]:
    """Perform concurrent searches to the CSDA STAC instance."""
    async for item in advanced.flatmap.raw(query, client.search, task_limit=task_limit):
        yield item


@pipable_operator
async def extract_links(item_collection: AsyncIterable[CSDAItemCollection], client: Client) -> AsyncIterator[DownloadLink]:
    """Extract links from STAC features."""
    base_url = client.config.api

    found: MutableMapping[str, bool] = LRUCache(maxsize=client.config.max_deduplication_cache)
    async with streamcontext(item_collection) as streamer:
        async for feature in streamer:
            for item in feature.features:
                for asset in item.assets.values():
                    file = asset.href.split("/")[-1]
                    if file in found:
                        continue
                    found[file] = True
                    yield DownloadLink.parse_url(f"{base_url}{asset.href.lstrip('/')}")


@pipable_operator
async def download(
    links: AsyncIterable[DownloadLink],
    client: Client,
    prefix: str,
    task_limit: Optional[int] = None,
) -> AsyncIterator[Optional[str]]:
    """Download links concurrently."""

    async def _download(
        *links: DownloadLink,
    ) -> AsyncIterable[Optional[str]]:
        for link in links:
            yield await client.download_file(url=link, prefix=prefix)

    async for success in advanced.flatmap.raw(links, _download, task_limit=task_limit):
        yield success



from __future__ import annotations

import logging

from pydantic import HttpUrl, SecretStr
from pydantic_settings import BaseSettings, SettingsConfigDict

logger = logging.getLogger()
_env_prefix = "csda_"


class Settings(BaseSettings):
    model_config = SettingsConfigDict(
        env_prefix=_env_prefix,
        env_file=".env",
        env_file_encoding="utf-8",
        extra="ignore",
    )

    username: str = ""
    password: SecretStr = SecretStr("")
    api: HttpUrl = HttpUrl("https://nasa-csda.wx.spire.com/")
    cognito_client_id: str = "7agre1j1gooj2jng6mkddasp9o"
    cognito_endpoint: str = "https://cognito-idp.us-east-1.amazonaws.com/"

    search_page_size: int = 100
    concurrent_downloads: int = 12
    item_buffer_size: int = 10
    retry_count: int = 10
    max_retry_wait_seconds: int = 30
    concurrent_searches: int = 4
    use_http2: bool = True
    download_progress: bool = False
    max_deduplication_cache: int = 1000



import logging

from httpx import AsyncHTTPTransport, Request, Response
from tenacity import AsyncRetrying, stop_after_attempt, wait_exponential_jitter

from nasa_csda.config import Settings

logger = logging.getLogger(__name__)


class RetryableTransport(AsyncHTTPTransport):
    def __init__(self, config: Settings, *args, **kwargs) -> None:
        kwargs.setdefault("http2", config.use_http2)
        super().__init__(*args, **kwargs)
        self.__config = config

    async def handle_async_request(self, request: Request) -> Response:
        path = request.url.path
        async for attempt in AsyncRetrying(
            stop=stop_after_attempt(self.__config.retry_count),
            wait=wait_exponential_jitter(max=self.__config.max_retry_wait_seconds),
            reraise=True,
        ):
            with attempt:
                resp = await super().handle_async_request(request)
                if resp.status_code >= 500:
                    msg = f"Failed request to /{path} ({resp.status_code}) attempt {attempt.retry_state.attempt_number}/10"
                    logger.debug(msg)
                    raise Exception(await resp.aread())
                return resp
        assert False



from datetime import datetime
from typing import AsyncIterator, TextIO, Optional, TypeVar

from aiostream import pipe, stream
import asyncclick as click
from pydantic import SecretStr, ValidationError
from tqdm.asyncio import tqdm
from tqdm.contrib.logging import logging_redirect_tqdm

from nasa_csda.buffer import Buffer
from nasa_csda.client import Client
from nasa_csda.config import Settings
from nasa_csda.models.bulk_download import BulkDownload
from nasa_csda.models.item_collection import CSDAItemCollection
from nasa_csda.models.search import CSDASearch
from nasa_csda.streaming import download, extract_links, search

T = TypeVar("T")


async def _run_search(
    client: Client,
    *queries: CSDASearch,
    mode: str,
    limit: Optional[int],
    destination: str,
    progress: bool,
) -> None:
    async def identity(x: T) -> T:
        return x

    async def split_queries(*qs: CSDASearch) -> AsyncIterator[CSDASearch]:
        for q in qs:
            for s in q.split():
                yield s

    async with client.session(), Buffer[CSDAItemCollection, [], CSDASearch](client.config.item_buffer_size) as buffered:
        p = (
            stream.iterate(queries)
            | pipe.flatmap(split_queries)
            | search.pipe(client, task_limit=client.config.concurrent_searches)
            | buffered.pipe()
        )
        if mode != "raw":
            p |= extract_links.pipe(client=client)  # type: ignore

        if mode == "download":
            p = p | download.pipe(client=client, prefix=destination, task_limit=10) | pipe.filter(identity)  # type: ignore

        if limit is not None:
            p = p | pipe.take(limit)

        with tqdm(unit=" files", disable=not progress) as pbar, logging_redirect_tqdm():
            async with p.stream() as streamer:
                async for item in streamer:
                    pbar.update(1)
                    if mode == "raw":
                        click.echo(item.model_dump_json())
                    elif mode == "list":
                        click.echo(item)


@click.group()
@click.option("--username", envvar="CSDA_USERNAME")
@click.option("--password", envvar="CSDA_PASSWORD")
@click.option("--settings-file", envvar="CSDA_SETTINGS_FILE", default=".env")
@click.pass_context
def cli(ctx, username, password, settings_file):
    kwargs = {}
    if username:
        kwargs["username"] = username
    if password:
        kwargs["password"] = password
    config = Settings(_env_file=settings_file, **kwargs)
    if not config.username or not config.password.get_secret_value():
        raise click.BadParameter(
            "username and password must be provided",
            ctx,
            None,
            "username, password",
        )
    if username is not None:
        config = config.model_copy(update={"username": username})
    if password is not None:
        config = config.model_copy(update={"password": SecretStr(password)})
    ctx.ensure_object(dict)
    ctx.obj["config"] = config
    ctx.obj["client"] = Client(config)


@cli.command()
@click.pass_obj
async def token(obj):
    """Print an authentication token for STAC endpoints.

    This will log into the STAC catalog with the provided credentials.  The
    token returned will be valid for up to one hour.  In order to use this
    token, pass it as a Bearer token in the Authentication header.

    \b
    For example
    curl -H "Authentication:Bearer <token>" https://nasa-csda.wx.spire.com/stac/collections
    """
    client: Client = obj["client"]
    async with client.session():
        click.echo(await client.get_token())


@cli.command()
@click.option("--start-date", type=click.DateTime(), default=datetime(2019, 1, 1))
@click.option("--end-date", type=click.DateTime(), default=datetime.now())
@click.option("--min-latitude", type=click.FloatRange(min=-90, max=90), default=-90)
@click.option("--max-latitude", type=click.FloatRange(min=-90, max=90), default=90)
@click.option("--min-longitude", type=click.FloatRange(min=-180, max=180), default=-180)
@click.option("--max-longitude", type=click.FloatRange(min=-180, max=180), default=180)
@click.option("--products", default="", help="A comma-separated list of product names")
@click.option("--limit", type=click.IntRange(min=1), help="Stop after finding this many items")
@click.option(
    "--mode",
    type=click.Choice(["download", "list", "raw"]),
    default="download",
    help="Operational mode (see help)",
)
@click.option(
    "--progress/--no-progress",
    default=True,
    help="Enable progress reporting while downloading",
)
@click.option(
    "--destination",
    default="csda/{product}/{datetime:%Y}/{datetime:%m}/{datetime:%d}",
    help="Configure download location",
)
@click.option(
    "--overwrite/--no-overwrite",
    default=True,
    help="Overwrite existing files when downloading",
)
@click.pass_context
async def query(
    ctx,
    start_date: datetime,
    end_date: datetime,
    min_latitude: float,
    max_latitude: float,
    min_longitude: float,
    max_longitude: float,
    products: str,
    limit: Optional[int],
    mode: str,
    progress: bool,
    destination: str,
    overwrite: bool,
):
    """
    Query data from Spire's CSDA catalog.

    This command can return data in three modes:

    \b
    * download: download files
    * list:     show urls
    * raw:      show raw geojson items

    When in download mode, you can configure the path where the file is placed
    with a format-style string as the `destination` parameter. This parameter
    allows you to provide any static path as well as sort the downloads into
    prefixes based on properties of the file.  These properties include:

    \b
    * collection
    * datetime
    * receiver
    * product

    In addition, it supports standard python format strings, so you can extract
    components of the datetime as path elements such as `{datetime:%Y-%m-%d}`
    to make a subdirectory for each day.  The path string is automatically
    converted to backslashes when running in Windows.

    The default destination is `csda/{product}/{datetime:%Y}/{datetime:%m}/{datetime:%d}`.
    """
    obj = ctx.obj
    if min_latitude > max_latitude:
        raise click.BadParameter(
            "min-latitude must be <= max-latitude",
            ctx,
            None,
            "min-latitude, max-latitude",
        )
    if min_longitude > max_longitude:
        raise click.BadParameter(
            "min-longitude must be <= max-longitude",
            ctx,
            None,
            "min-longitude, max-longitude",
        )
    if start_date > end_date:
        raise click.BadParameter(
            "start-date must be <= end-date",
            ctx,
            None,
            "start-date, end-date",
        )

    client: Client = obj["client"]
    query = CSDASearch.build_query(
        start_date,
        end_date,
        min_latitude,
        max_latitude,
        min_longitude,
        max_longitude,
        products,
    )

    await _run_search(client, query, mode=mode, limit=limit, destination=destination, progress=progress)


@cli.command()
@click.option("--limit", type=click.IntRange(min=1), help="Stop after downloading this many files")
@click.option(
    "--progress/--no-progress",
    default=True,
    help="Enable progress reporting while downloading",
)
@click.option(
    "--destination",
    default="csda/{product}/{datetime:%Y}/{datetime:%m}/{datetime:%d}",
    help="Configure download location",
)
@click.option(
    "--overwrite/--no-overwrite",
    default=True,
    help="Overwrite existing files when downloading",
)
@click.argument("config", type=click.File("r"), nargs=-1)
@click.pass_context
async def bulk_download(
    ctx,
    limit: Optional[int],
    progress: bool,
    destination: str,
    overwrite: bool,
    config: list[TextIO],
):
    """
    Bulk download files the Spire's CSDA catalog.

    You can configure the path where the file is placed with a format-style
    string as the `destination` parameter. This parameter allows you to provide
    any static path as well as sort the downloads into prefixes based on
    properties of the file.  These properties include:

    \b
    * collection
    * datetime
    * receiver
    * product

    In addition, it supports standard python format strings, so you can extract
    components of the datetime as path elements such as `{datetime:%Y-%m-%d}`
    to make a subdirectory for each day.  The path string is automatically
    converted to backslashes when running in Windows.

    The default destination is `csda/{product}/{datetime:%Y}/{datetime:%m}/{datetime:%d}`.
    """
    client: Client = ctx.obj["client"]
    for file in config:
        try:
            queries = BulkDownload.model_validate_json(file.read()).queries
        except ValidationError as err:
            click.secho(str(err), err=True, fg="red")
            raise click.BadParameter(
                "could not parse file",
                ctx,
                None,
                "config",
            )

    await _run_search(client, *queries, mode="download", limit=limit, destination=destination, progress=progress)



from typing import Optional

from stac_pydantic.api.item_collection import ItemCollection
from stac_pydantic.api.links import PaginationLink, SearchLink


class CSDAItemCollection(ItemCollection):
    @property
    def next_token(self) -> Optional[str]:
        if self.links is None:
            return None
        for link in self.links.link_iterator():
            if not isinstance(link, (PaginationLink, SearchLink)):
                continue
            rel = link.rel
            if rel == "next" and link.body is not None:
                return link.body.get("token")
        return None



from asyncio import CancelledError, create_task, Queue, Task
from contextlib import AbstractAsyncContextManager
import logging
from typing import Any, AsyncIterable, AsyncIterator, Generic, Optional, TypeVar

from aiostream.core import pipable_operator, PipableOperator, streamcontext
from typing_extensions import ParamSpec

A = TypeVar("A", contravariant=True)
P = ParamSpec("P")
T = TypeVar("T", covariant=True)

logger = logging.getLogger(__name__)


class Buffer(AbstractAsyncContextManager, Generic[A, P, T]):
    """Create a buffering context manager..

    >>> async with Buffer() as buffered:
            # the stream `p` will be iterated in the background
            p |= buffered.pipe()
            async with streamcontext(p) as streamer:
                async for item in streamer:
                    item
    """

    def __init__(self, size: int = 1000) -> None:
        self.queue: Queue = Queue(size)
        self.task: Optional[Task] = None

    @staticmethod
    async def _producer(it: AsyncIterable[Any], queue: Queue) -> None:
        async with streamcontext(it) as items:
            try:
                async for item in items:
                    await queue.put((False, item))
            except BaseException as err:
                await queue.put((True, err))
            else:
                await queue.put((None, None))

    async def __aenter__(self) -> PipableOperator[A, P, T]:

        @pipable_operator
        async def buffered(it: AsyncIterable[A], *args: P.args, **kwargs: P.kwargs) -> AsyncIterator[T]:
            """Buffer values from an iterator in the background.

            This function creates an asynchronous task that pushes data into a
            bounded queue.
            """
            self.task = create_task(self._producer(it, self.queue))
            try:
                while True:
                    status, item = await self.queue.get()
                    if status:
                        raise item
                    if status is None:
                        break
                    yield item
            except CancelledError:
                pass

        return buffered

    async def __aexit__(self, *args):
        if self.task:
            self.task.cancel()
            await self.task
